# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from ctypes import c_buffer
from .ktx_error_code import KtxErrorCode, KtxError
from pyktx.native import ffi, lib
from typing import Dict, Optional


class KtxHashList:
    """Opaque handle to a ktxHashList implemented in C."""

    def __init__(self, ptr):
        self._ptr = ptr

    def add_kv_pair(self, key: str, value: bytes) -> None:
        """Add a key value pair to a hash list."""

        error = lib.ktxHashList_AddKVPair(self._ptr, key.encode('ascii'), len(value), value)

        if int(error) != KtxErrorCode.SUCCESS:
            raise KtxError('ktxHashList_AddKVPair', KtxErrorCode(error))

    def delete_kv_pair(self, key: str) -> None:
        """Delete a key value pair in a hash list."""

        error = lib.ktxHashList_DeleteKVPair(self._ptr, key.encode('ascii'))

        if int(error) != KtxErrorCode.SUCCESS:
            raise KtxError('ktxHashList_DeleteKVPair', KtxErrorCode(error))

    def find_value(self, key: str) -> Optional[c_buffer]:
        """Looks up a key in a hash list and returns the value."""

        data = lib.PY_ktxHashList_FindValue(self._ptr, key.encode('ascii'))

        if data.error == KtxErrorCode.NOT_FOUND:
            return None
        if data.error != KtxErrorCode.SUCCESS:
            raise KtxError('ktxHashList_FindValue', KtxErrorCode(data.error))

        return ffi.buffer(data.bytes, data.size)

    def copy(self) -> Dict[str, bytes]:
        """Copy the hash list into a dict. This is recommended if you reading the hash list."""

        kv_data = {}
        entry = lib.PY_ktxHashList_get_listHead(self._ptr)

        while entry != ffi.NULL:
            key_data = lib.PY_ktxHashListEntry_GetKey(entry)
            value_data = lib.PY_ktxHashListEntry_GetValue(entry)

            if int(key_data.error) != KtxErrorCode.SUCCESS:
                raise KtxError('ktxHashListEntry_GetKey', KtxErrorCode(key_data.error))
            if int(value_data.error) != KtxErrorCode.SUCCESS:
                raise KtxError('ktxHashListEntry_GetValue', KtxErrorCode(value_data.error))

            key: str = ffi.buffer(key_data.bytes, key_data.size)[:].decode(encoding='utf-8')
            value = ffi.buffer(value_data.bytes, value_data.size)[:]

            if key.endswith('\x00'):
                key = key[:-1]

            kv_data[key] = value

            entry = lib.ktxHashList_Next(entry)

        return kv_data



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from .ktx_pack_uastc_flag_bits import KtxPackUastcFlagBits


@dataclass
class KtxBasisParams:
    """Data for passing extended params to KtxTexture2.compressBasis()."""

    uastc: bool = False
    """True to use UASTC base, false to use ETC1S base."""

    verbose: bool = False
    """If true, prints Basis Universal encoder operation details to stdout. Not recommended for GUI apps."""

    no_sse: bool = False
    """True to forbid use of the SSE instruction set. Ignored if CPU does not support SSE."""

    thread_count: int = 1
    """Number of threads used for compression. Default is 1."""

    compression_level: int = 0
    """
    Encoding speed vs. quality tradeoff. Range is [0,5].

    Higher values are slower, but give higher quality. There is no default.
    Callers must explicitly set this value. Callers can use
    KTX_ETC1S_DEFAULT_COMPRESSION_LEVEL as a default value. Currently this is 2.
    """

    quality_level: int = 0
    """
    Compression quality. Range is [1,255].

    Lower gives better compression/lower quality/faster.
    Higher gives less compression/higher quality/slower.
    This automatically determines values for max_endpoints, max_selectors,
    endpoint_rdo_threshold, and selector_rdo_threshold for the target quality
    level. Setting these parameters overrides the values determined by quality_level
    which defaults to 128 if neither it nor both of max_endpoints and max_selectors
    have been set.

    Both of max_endpoints and max_selectors must be set for them to have any effect.
    quality_level will only determine values for endpoint_rdo_threshold and selector_rdo_threshold
    when its value exceeds 128, otherwise their defaults will be used.
    """

    max_endpoints: int = 0
    """
    Manually set the max number of color endpoint clusters.

    Range is [1,16128]. Default is 0, unset. If this is set, max_selectors
    must also be set, otherwise the value will be ignored.
    """

    endpoint_rdo_threshold: int = 0
    """
    Set endpoint RDO quality threshold. The default is 1.25.

    Lower is higher quality but less quality per output bit (try [1.0,3.0].
    This will override the value chosen by quality_level.
    """

    max_selectors: int = 0
    """
    Manually set the max number of color selector clusters. Range is [1,16128].

    Default is 0, unset. If this is set, max_endpoints must also be set, otherwise
    the value will be ignored.
    """

    selector_rdo_threshold: int = 0
    """
    Set selector RDO quality threshold. The default is 1.5.

    Lower is higher quality but less quality per output bit (try [1.0,3.0]).
    This will override the value chosen by @c qualityLevel.
    """

    input_swizzle: bytes = bytes(4)
    """
    A swizzle to apply before encoding.

    It must match the regular expression /^[rgba01]{4}$/. If both this and
    pre_swizzle are specified KtxTexture2.compressBasis() will raise INVALID_OPERATION.
    """

    normal_map: bool = False
    """
    Tunes codec parameters for better quality on normal maps (no
    selector RDO, no endpoint RDO) and sets the texture's DFD appropriately.

    Only valid for linear textures.
    """

    pre_swizzle: bool = False
    """
    If the texture has swizzle metadata, apply it before compressing.

    Swizzling, like rabb may yield drastically different error metrics
    if done after supercompression.
    """

    separate_rg_to_rgb_a: bool = False
    """
    This was and is a no-op.

    2-component inputs have always been automatically separated
    using an "rrrg" input_swizzle.
    """

    no_endpoint_rdo: bool = False
    """
    Disable endpoint rate distortion optimizations.

    Slightly faster, less noisy output, but lower quality per output bit.
    """

    no_selector_rdo: bool = False
    """
    Disable selector rate distortion optimizations.

    Slightly faster, less noisy output, but lower quality per output bit.
    """

    uastc_flags: int = KtxPackUastcFlagBits.FASTEST
    """
    A set of KtxPackUastcFlagBits controlling UASTC encoding.

    The most important value is the level given in the
    least-significant 4 bits which selects a speed vs quality tradeoff.
    """

    uastc_rdo: bool = False
    """Enable Rate Distortion Optimization (RDO) post-processing."""

    uastc_rdo_quality_scalar: float = 0.
    """
    UASTC RDO quality scalar (lambda).

    Lower values yield higher quality/larger LZ compressed files, higher
    values yield lower quality/smaller LZ compressed files. A good range to
    try is [.2,4]. Full range is [.001,50.0]. Default is 1.0.
    """

    uastc_rdo_dict_size: int = 0
    """
    UASTC RDO dictionary size in bytes. Default is 4096. Lower
    values=faster, but give less compression. Range is [64,65536].
    """

    uastc_rdo_max_smooth_block_error_scale: float = 10.
    """
    UASTC RDO max smooth block error scale. Range is [1,300].
    Default is 10.0, 1.0 is disabled. Larger values suppress more
    artifacts (and allocate more bits) on smooth blocks.
    """

    uastc_rdo_max_smooth_block_std_dev: float = 18.
    """
    UASTC RDO max smooth block standard deviation. Range is
    [.01,65536.0]. Default is 18.0. Larger values expand the range of
    blocks considered smooth.
    """

    uastc_rdo_dont_favor_simpler_modes: bool = False
    """Do not favor simpler UASTC modes in RDO mode."""

    uastc_rdo_no_multithreading: bool = False
    """Disable RDO multithreading (slightly higher compression, deterministic)."""



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from enum import IntEnum


class KtxPackUastcFlagBits(IntEnum):
    """Flags specifiying UASTC encoding options."""

    FASTEST = 0
    """Fastest compression. 43.45dB."""

    FASTER = 1
    """Faster compression. 46.49dB."""

    DEFAULT = 2
    """Default compression. 47.47dB."""

    SLOWER = 3
    """Slower compression. 48.01dB."""

    VERY_SLOW = 4
    """Very slow compression. 48.24dB."""

    MAX_LEVEL = VERY_SLOW
    """Maximum supported quality level."""

    LEVEL_MASK = 0xF
    """Mask to extract the level from the other bits."""

    FAVOR_UASTC_ERROR = 8
    """Optimize for lowest UASTC error."""

    FAVOR_BC7_ERROR = 16
    """Optimize for lowest BC7 error."""

    ETC1_FASTER_HINTS = 64
    """Optimize for faster transcoding to ETC1."""

    ETC1_FASTEST_HINTS = 128
    """Optimize for fastest transcoding to ETC1."""

    ETC1_DISABLE_FLIP_AND_INDIVIDUAL = 256
    """Not documented in BasisU code."""



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from enum import IntEnum


class KtxPackAstcQualityLevels(IntEnum):
    """Options specifiying ASTC encoding quality levels."""

    FASTEST = 0
    """Fastest compression."""

    FAST = 10
    """Fast compression."""

    MEDIUM = 60
    """Medium compression."""

    THOROUGH = 98
    """Slower compression."""

    EXHAUSTIVE = 100
    """Very slow compression."""

    MAX = EXHAUSTIVE
    """Maximum supported quality level."""



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from enum import IntEnum


class KtxTextureCreateStorage(IntEnum):
    """Enum for requesting, or not, allocation of storage for images."""

    NO = 0
    """Don't allocate any image storage."""

    ALLOC = 1
    """Allocate image storage."""

